

#include "dosdrvr.h"
#include "header.h"
#include "utility.h"
#include "irq.h"
#include "scsi.h"
#include "ncr_400.h"


/* Register offsets for ReadReg() and WriteReg() */
#define CSD		0x3880
#define ODR		0x3880
#define ICR		0x3881
#define MR		0x3882
#define TCR		0x3883
#define CSC		0x3884
#define SER		0x3884
#define ISR		0x3885
#define SDS		0x3885
#define IDR		0x3886
#define SDT		0x3886
#define RIR		0x3887
#define SDIR	0x3887
#define CR0		0x3980
#define BCR		0x3981
#define SR		0x3982
#define CR1		0x3983
#define CR2		0x3984
#define IAR		0x3985	
#define IDX		0x3986	
#define TR		0x3987	

// Interrupt Status Register (ISR) bits
#define EDMA	0x80
#define DRQ		0x40
#define PE		0x20
#define IRQ		0x10
#define PHM		0x08
#define BSYE	0x04
#define SATN	0x02
#define SACK	0x01

// Initiator Command Register (ICR) bits
#define RST		0x80
#define ARB		0x40
#define LARB	0x20
#define ACK		0x10
#define BSY		0x08
#define SEL		0x04
#define ATN		0x02
#define DB		0x01

// Current SCSI Control Bits 
#define SRST	0x80
#define SBSY	0x40
#define SREQ	0x20
#define SMSG	0x10
#define SCD		0x08
#define SIO		0x04
#define SSEL	0x02
#define SDBP	0x01

// Target Command Register (TCR) bits
#define LBS		0x80
#define REQ		0x08
#define MSG		0x04
#define CD		0x02
#define IO		0x01



// SCSI bus phases (sent to TCR register)
#define PHASE_DATA_OUT	(0)
#define PHASE_DATA_IN	(IO)
#define PHASE_MSG_OUT	(MSG | CD)
#define PHASE_MSG_IN	(MSG | CD | IO)
#define PHASE_CMD		(CD)
#define PHASE_STATUS	(CD | IO)



#define NCR_FP(s,o)		((void FAR *)(((unsigned long)(s) << 16) | (o)))

#define NcrHostBuffer	((unsigned char FAR *)NCR_FP(BaseSegment,0x3900))
#define NcrScratchRam	((unsigned char FAR *)NCR_FP(BaseSegment,0x3800))


static unsigned	BaseSegment;
static int		IrqLevel;
static int		ScsiId;
static int		ScsiIdBit;
static int		IrqInstalled;

static unsigned MaxAckDelay = 0x7FFF;	// raw loop counter
static unsigned MaxEopDelay = 0x7FFF;	// raw loop counter
static unsigned MaxSelDelay = 0x7FFF;	// about 8 us per (on 8MHz ISA bus)
static unsigned MaxRegDelay = 0x7FFF;	// raw loop counter

static unsigned MaxArbRetries = 16;

static int		LastError;
static int		LastTCR;

static int TcrPhaseMap[8] = 
	{
	0,						// SCSI_PHASE_BUS_FREE	0
	CD,						// SCSI_PHASE_CMD		1
	CD | IO,				// SCSI_PHASE_STATUS	2
	MSG | CD,				// SCSI_PHASE_MSG_OUT	3
	MSG | CD | IO,			// SCSI_PHASE_MSG_IN	4
	0,						// SCSI_PHASE_DATA_OUT	6
	IO						// SCSI_PHASE_DATA_IN	5
	};

	
static void WriteReg( unsigned reg, int val )
	{
	*(unsigned char FAR *)NCR_FP(BaseSegment,(reg)) = val;
	}

static unsigned char ReadReg( unsigned reg )
	{
	return *(unsigned char FAR *)NCR_FP(BaseSegment,(reg));
	}

static void EnableTargetMode()
	{
	WriteReg(MR,0x40);			// Target-mode operation
	WriteReg(SER,ScsiIdBit);	// Set target SCSI ID bit
	}

int adapter_GetScsiId()
	{
	return ScsiId;
	}

int adapter_GetScsiBits()
	{
	return ((ReadReg(ISR) & 3) << 8) | ReadReg(CSC);
	}

static void SelectPhase(int phase)
	{
	// Save the current phase (used in other routines)
	LastTCR = TcrPhaseMap[phase];

	// If we are now in a phase that allows us
	// to send data to the initiator we must
	// also enable the data bus drivers.  If not,
	// disable the data bus drivers to prevent
	// contention with the initiator.

	if (LastTCR & IO)
		{
		WriteReg(TCR,LastTCR);	// Change phase first to avoid contention
		WriteReg(ICR,BSY|DB);
		}
	else
		{
		WriteReg(ICR,BSY);		// Disable drivers first to avoid contention
		WriteReg(TCR,LastTCR);
		}
	}


static void ClearIrq(void)
	{
	ReadReg(RIR);		// reading this register resets the interrupt 
	ReadReg(CR0);
	if (IrqLevel > 0)
		SendEOI(IrqLevel);
	}


// Wait for ACK from the initiator

static int WaitForAck()
	{
	unsigned delay = 0;

	while ( (ReadReg(ISR) & SACK) == 0 )
		{
		if (delay++ >= MaxAckDelay)
			return 0;
		}

	return 1;
	}



// Wait for access to 53C400 registers

static int WaitForRegAccess()
	{
	unsigned delay = 0;
	while ( (ReadReg(CR0) & 0x80) == 0 )
		{
		if (delay++ > MaxRegDelay)
			return 0;
		}
	return 1;
	}



static int WriteBlocks(char FAR *bufp, int nblocks)
	{
	int i,n;
	int nwritten = 0;
	int error = 0;
	unsigned delay;

	while ((nwritten < nblocks) && !error)
		{
		n = (nblocks < 256) ? nblocks : 256;
		WriteReg(MR,0x6A);
		WriteReg(CR0,ReadReg(CR0) & ~0xC8);
		WriteReg(SDT,0xFF);		
		WriteReg(BCR,n);
		for (i=0; i<n; i++)
			{
			// Wait for 53C400 to read an entire 128 byte buffer
			while ((ReadReg(CR0) & 0x05) == 0x04)
				;
			// End of process interrupt (problem with transfer)?
			if ( ReadReg(CR0) & 0x01 )
				{
				error = 1;
				break;
				}
			_fmemcpy(NcrHostBuffer,bufp,128);
			bufp += 128;
			nwritten += 1;
			}
		}

	// Wait for end of process interrupt
	delay = 0;
	while ( (ReadReg(CR0) & 0x01) == 0 )
		{
		if (delay++ > MaxEopDelay)
			break;
		}

	WaitForRegAccess();

	WriteReg(MR,ReadReg(MR) & ~0x2A);	// reset mode
	ReadReg(RIR);						// clear IRQ

	return nwritten;
	}

static int ReadBlocks(char FAR *bufp, int nblocks)
	{
	int i,n;
	int nread = 0;
	int error = 0;
	unsigned delay;

	while ((nread < nblocks) && (!error) )
		{
		n = (nblocks < 256) ? nblocks : 256;
		WriteReg(MR,0x6A);
		WriteReg(CR0,(ReadReg(CR0) & ~0xC8) | 0x40);
		WriteReg(SDT,0xFF);		
		WriteReg(BCR,n);
		for (i=0; i<n; i++)
			{
			// Wait for 53C400 to read an entire 128 byte buffer
			while ((ReadReg(CR0) & 0x05) == 0x04)
				;
			// End of process interrupt (problem with transfer)?
			if ( ReadReg(CR0) & 0x01 )
				{
				error = 1;
				break;
				}
			_fmemcpy(bufp,NcrHostBuffer,128);
			bufp += 128;
			nread += 1;
			}
		}

	// Wait for end of process interrupt
	delay = 0;
	while ( (ReadReg(CR0) & 0x01) == 0 )
		{
		if (delay++ > MaxEopDelay)
			break;
		}


	WaitForRegAccess();

	WriteReg(MR,ReadReg(MR) & ~0x2A);	// reset mode
	WriteReg(CR0,ReadReg(CR0) & ~0xC8);	// clear direction bit
	ReadReg(RIR);						// clear IRQ

	return nread;
	}

int adapter_Read(int phase, char FAR *bufp, int maxbytes)
	{
	int nread = 0;

	SelectPhase(phase);

	if ( (phase == SCSI_PHASE_DATA_OUT) && (maxbytes > 128) )
		{
		// Use fast block-mode read
		int n = maxbytes / 128;
		nread = ReadBlocks(bufp,n);
		if (nread != n)
			return nread * 128;
		nread *= 128;
		bufp += nread;
		}

	while (nread < maxbytes)
		{
		WriteReg(TCR, LastTCR | REQ);
		if (!WaitForAck())
			{
			// Error
			WriteReg(TCR, LastTCR);
			return nread;
			}
		*bufp++ = ReadReg(CSD);
		WriteReg(TCR, LastTCR);
		nread++;
		}

	return nread;
	}

int adapter_Write(int phase, char FAR *bufp, int nbytes)
	{
	int nwritten = 0;

	SelectPhase(phase);

	if ( (phase == SCSI_PHASE_DATA_IN) && (nbytes > 128) )
		{
		// Use fast block-mode write
		int n = nbytes / 128;
		nwritten = WriteBlocks(bufp,n);
		if (nwritten != n)
			return nwritten * 128;
		nwritten *= 128;
		bufp += nwritten;
		}

	while (nwritten < nbytes)
		{
		WriteReg(ODR,*bufp++);
		WriteReg(TCR, LastTCR | REQ);
		if (!WaitForAck())
			{
			WriteReg(TCR, LastTCR);
			return nwritten;
			}
		WriteReg(TCR, LastTCR);
		nwritten++;
		}

	return nwritten;
	}

void adapter_Disconnect(void)
	{
	SelectPhase(SCSI_PHASE_BUS_FREE);
	WriteReg(ICR,0);
	WriteReg(MR,0x40);					// Enable target mode
	WriteReg(SER,ScsiIdBit);			// Enable target selection
	}

int adapter_Reconnect(int initiator_id)
	{
	unsigned tries;
	unsigned i;
	unsigned char x;

	for (tries=0; tries<=MaxArbRetries; i++)
		{
		WriteReg(SER,0);					// disable target selection
		WriteReg(ODR,ScsiIdBit);			// set our SCSI ID for arbitration
		WriteReg(MR,0x41);					// Start arbitration process

		while (ReadReg(ICR) & 0x40)			// ARB in progress
			;

		// Kill at least 2.2 us (arb delay) (assumes 8MHz bus)
		for (i=0; i<10; i++)
			WriteReg(IDX,0);

		if (ReadReg(ICR) & 0x20)			// Lost ARB bit
			continue;

		// Check for higher priority device
		x = ReadReg(CSD) & ~ScsiIdBit;
		if (x < ScsiIdBit)
			{
			// We've won the arbitration
			WriteReg(ICR,SEL);		// Set SEL to enter re-selection phase
			// Kill at least 1.2 us (bus clear and settle) (assumes 8MHz bus)
			for (i=0; i<6; i++)
				WriteReg(IDX,0);
			WriteReg(TCR,IO);		// Set IO to become the target
			WriteReg(ODR,ScsiIdBit | (1 << initiator_id));
			WriteReg(ICR,SEL|BSY|DB);	// Set BSY and DB
			WriteReg(MR,0x40);			// Clear ARB bit
			WriteReg(ICR,SEL|DB);		// Clear BSY

			for (i=0; i<MaxSelDelay; i++)
				{
				if (ReadReg(CSC) & SBSY)
					{
					WriteReg(ICR,SEL|BSY|DB);	// Set BSY and DB
					for (i=0; i<1; i++)			// Wait two deskew delays
						WriteReg(IDX,0);
					WriteReg(ICR,BSY);			// Clear SEL and DB
					return 1;
					}
				// Add a bit more time to the delay
				for (x=0; x<16; x)			// Wait about 8 microseconds
					WriteReg(IDX,0);		// to make the delay work right
				}

			// Reselection timeout
			}
		WriteReg(TCR,0);
		WriteReg(ICR,0);
		WriteReg(MR,0x40);			// Clear ARB bit
		for (i=0; i<10; i++)
			WriteReg(IDX,0);
		}

	// Go back to target mode and return error
	WriteReg(MR,0x40);					// Enable target mode
	WriteReg(SER,ScsiIdBit);			// Enable target selection
	return 0;
	}


void adapter_IrqHandler()
	{
	unsigned i,delay;
	unsigned char tmp;

	// First check for a SCSI bus reset signal
	// If detected we have to put the 53C400 
	// back into target mode (cleared by reset)

	if (ReadReg(CSC) & SRST)			// Reset?
		{
		EnableTargetMode();
		scsi_BusResetDetected();
		ClearIrq();
		return;
		}

	if (ReadReg(ISR) & PE)				// Parity error?
		{
		ClearIrq();
		return;
		}

	if (ReadReg(CSC) & SSEL)			// Selection?
		{
		// Make sure we are properly selected 
		// by checking our ID bit on the bus.  
		tmp = ReadReg(CSD);
		if (tmp & ScsiIdBit)
			{
			// We are selected, find the initiator ID
			tmp &= ~ScsiIdBit;			// clear our ID bit
			for (i=0; i<8; i++)
				{
				if (tmp & 1)
					{
					// We have a valid selection, set BSY
					// and wait for initiator to drop SEL
					WriteReg(ICR,BSY);
					delay = 0;
					while (ReadReg(CSC) & SSEL)
						{
						if (delay++ > MaxSelDelay)
							{
							WriteReg(ICR,0);
							ClearIrq();
							return;
							}
						}
					// We are now in control of the SCSI bus

					WriteReg(CR0,0);		// disable 53C400 interrupts
					ClearIrq();				// clear this one
					_enable();				// allow other interrupt sources

					scsi_SelectionDetected(i,ScsiId);

					_disable();				// disable all interrupts
					WriteReg(CR0,0x10);		// allow 53C400 interrupts
					return;
					}
				tmp >>= 1;
				}
			}
		}

	// Spurious interrupt, clear it and return
	ClearIrq();
	return;
	}



int adapter_Init(char FAR *cmdline)
	{
	char FAR *s;
	int id;

	/* Extract segment address of card's register space */
	s = FindArg(cmdline,'S');
	if (s)
		BaseSegment = str_to_hex(s);
	else
		BaseSegment = 0xD800;

	/* Extract IRQ level */
	s = FindArg(cmdline,'I');
	if (s)
		{
		IrqLevel = str_to_dec(s);
		if (IrqLevel == 2)
			IrqLevel = 9;
		if (IrqLevel < 3 || IrqLevel > 15)
			IrqLevel = -1;
		}
	else
		IrqLevel = -1;					/* Disabled */

	/* Extract SCSI ID */
	s = FindArg(cmdline,'T');
	if (s)
		ScsiId = str_to_dec(s) & 0x07;
	else
		ScsiId = 0;
	ScsiIdBit = 1 << ScsiId;

	PutString("53C400 driver installing at ");  PutHexWord(BaseSegment);
	PutString(":0000, IRQ "); 
	if (IrqLevel >= 0)
		PutChar('0' + IrqLevel);
	else
		PutString("Disabled");
	PutString(", SCSI ID "); PutChar('0' + ScsiId);
	PutString("\r\n");

	WriteReg(CR0,0x80);				// Reset the 53C400
	WaitForRegAccess();
	WriteReg(CR0,0x00);
	if (ReadReg(CR0) & 0x80)		// should be set (register access)
		{
		// Card is responding
		WriteReg(CR0,0x80);			// Reset it again
		WaitForRegAccess();
		ReadReg(RIR);				// Reset IRQ
		WriteReg(MR,0x40);			// Target-mode operation
		WriteReg(SER,ScsiIdBit);	// Set target SCSI ID bit
		if (IrqLevel >= 0)
			{
			InstallIrqHandler(IrqLevel,adapter_IrqHandler);
			WriteReg(CR0,0x10);			// Enable IRQ on the 53C400
			ReadReg(RIR);				// Reset IRQ on the 53C400
			SendEOI(IrqLevel);			// Clear pending 8259 IRQ request
			EnableIrq(IrqLevel);		// Enable IRQ on the 8259
			IrqInstalled = 1;
			}
		return 1;
		}
	else
		{
		PutString("Card not responding!\r\n");
		return 0;
		}

	}


void adapter_Term()
	{
	if (IrqInstalled)
		{
		DisableIrq(IrqLevel);
		RemoveIrqHandler(IrqLevel);
		IrqInstalled = 0;
		}
	WriteReg(CR0,0x80);			// Reset the 53C400
	}



// This routine is called repeatedly while the TSPI manager
// is waiting for an interrupt.  It may also be called during
// a timer interrupt, so be sure to obey all interrupt-time 
// restrictions.

void adapter_Poll()
	{
	if (ReadReg(ISR) & 0x10)		// interrupt pending?
		{
		_disable();
		if (ReadReg(ISR) & 0x10)		// interrupt pending?
			{
			WriteReg(CR0,0x10);			// Enable IRQ on the 53C400
			adapter_IrqHandler();		// handle it
			WriteReg(CR0,0x10);			// Enable IRQ on the 53C400
			}
		_enable();
		}
	}


void adapter_Reset()
	{
	WriteReg(CR0,0x80);			// Reset it again
	WaitForRegAccess();
	ReadReg(RIR);				// Reset IRQ
	WriteReg(MR,0x40);			// Target-mode operation
	WriteReg(SER,ScsiIdBit);	// Set target SCSI ID bit
	if (IrqLevel >= 0)
		{
		WriteReg(CR0,0x10);			// Enable IRQ on the 53C400
		ClearIrq();
		EnableIrq(IrqLevel);		// Enable IRQ on the 8259
		}
	}
