

/*********************************************************
* SCSI.C
* 
* This module implements the SCSI BUS I/O routines 
* required for the TSPI manager. 
* 
* It works closely with the NCR_400 module to control
* the SCSI bus.  It is kept separate to simplify ports
* to SCSI adapters that work like the 53C400.
* 
* SCSI adapters with radically different interfaces 
* may require substantial modifications to this module.
*
*********************************************************/

#include <limits.h>
#include "dosdrvr.h"
#include "header.h"
#include "ncr_400.h"
#include "scsi.h"
#include "tspi.h"
#include "tspilink.h"



// This routine is called while the TSPI manager is waiting
// for an incoming SCSI event.  It gives non-IRQ capable
// adapters a chance to poll the SCSI bus for selections.

void scsi_Poll()
	{
	adapter_Poll();
	}



// This routine returns the adapter's SCSI ID

int scsi_GetScsiId(void)
	{
	return adapter_GetScsiId();
	}


// This routine is called to check the state of the ATN signal
int scsi_IsAtnAsserted(void)
	{
	return (adapter_GetScsiBits() & SCSI_BIT_ATN) ? 1 : 0;
	}


// This routine is called by the NCR_400 module whenever a SCSI
// bus reset is detected.

void scsi_BusResetDetected(void)
	{
	tspi_BusReset();
	}


// This routine completes the processing of a SCSI command.
// It sends the status byte, the COMMAND COMPLETE message,
// and disconnects from the SCSI bus.

int scsi_CompleteCommand(unsigned char status)
	{
	unsigned char msg_byte = 0x00;	// command complete
	adapter_Write(SCSI_PHASE_STATUS,&status,1);
	adapter_Write(SCSI_PHASE_MSG_IN,&msg_byte,1);
	adapter_Disconnect();
	return 1;
	}


// This routine transfers data from the user buffer to the initiator
// This is done in a DATA IN bus phase

int scsi_WriteToHost(void FAR *data_buffer, int nbytes)
	{
	char FAR *bufp = (char FAR *)data_buffer;
	int nwritten = 0;
	int n,m;

	while (nwritten < nbytes)
		{
		if (nbytes < 4096)
			m = nbytes;
		else
			m = 4096;
		n = adapter_Write(SCSI_PHASE_DATA_IN,bufp+nwritten,m);
		if (n < 0)
			nwritten += -n;
		else
			nwritten += n;
		if (n != m)
			break;
		}

	return nwritten;
	}

// This routine transfers data from the initiator to a user buffer
// This is done in a DATA OUT bus phase.

int scsi_ReadFromHost(void FAR *data_buffer, int nbytes)
	{
	char FAR *bufp = (char FAR *)data_buffer;
	int nread = 0;
	int n,m;

	while (nread < nbytes)
		{
		if (nbytes < 4096)
			m = nbytes;
		else
			m = 4096;
		n = adapter_Read(SCSI_PHASE_DATA_OUT,bufp+nread,m);
		if (n < 0)
			nread += -n;
		else
			nread += n;
		if (n != m)
			break;
		}

	return nread;
	}


// This routine transfers SCSI message bytes to the initiator.

int scsi_WriteMessageToHost(void FAR *msg_buffer, int nbytes)
	{
	return adapter_Write(SCSI_PHASE_MSG_IN,msg_buffer,nbytes);
	}


// This routine retrieves a single SCSI message from the initiator.

int scsi_ReadMessageFromHost(void FAR *msg_buffer, int maxbytes)
	{
	int err = 0;
	int n, nread = 0;
	unsigned char FAR *bufp = (unsigned char FAR *)msg_buffer;

	if (maxbytes < 1)
		return 0;

	if (adapter_Read(SCSI_PHASE_MSG_OUT,bufp,1) != 1)
		err = 1;
	else
		{
		nread++;
		if (maxbytes >= 2)
			{
			if ((*bufp & 0xF0) == 0x20)	// 2-byte message?
				{
				if (adapter_Read(SCSI_PHASE_MSG_OUT,bufp+1,1) != 1)
					err = 1;
				else
					nread++;
				}
			else 
				{
				if (*bufp == 0x01)			// extended message?
					{
					if (adapter_Read(SCSI_PHASE_MSG_OUT,bufp+1,1) != 1)
						err = 1;
					else
						{
						// Now get that many bytes (0 means 256 bytes)
						nread++;
						n = (*(bufp+1) == 0) ? 256 : *(bufp+1);
						if (maxbytes < 2+n)
							n = maxbytes - 2;
						if (n > 0)
							{
							if (adapter_Read(SCSI_PHASE_MSG_OUT,bufp+2,n) != n)
								err = 1;
							else
								nread += n;
							}
						}
					}
				}
			}
		}

	if (err)
		return -nread;
	}


// This routine disconnects the adapter from the SCSI bus, optionally
// sending a SAVE DATA POINTER message

int scsi_Disconnect(int save_data_pointer)
	{
	unsigned char msg;

	if (save_data_pointer)
		{
		msg = 0x02;								// SAVE DATA POINTER message
		adapter_Write(SCSI_PHASE_MSG_IN,&msg,1);
		}
		
	msg = 0x04;									// DISCONNECT message
	adapter_Write(SCSI_PHASE_MSG_IN,&msg,1);
	adapter_Disconnect();
	return 1;
	}


// This routine reconnects the adapter to the specified initiator

int scsi_Reconnect(int initiator_id, unsigned char identify_msg)
	{
	// Ignore for now
	if (adapter_Reconnect(initiator_id))
		{
		if (identify_msg & 0x80)			// valid identify message?
			adapter_Write(SCSI_PHASE_MSG_IN,&identify_msg,1);
		}
	return 0;
	}


// This routine is called from the NCR_400 module whenever we
// are the target of a valid selection.  It processes any message
// bytes waiting at the host, gets the CDB from the host, and forwards
// it to the TSPI manager for processing.

void scsi_SelectionDetected( int initiator_id, int target_id )
	{
	unsigned char identify_msg = 0;
	int lun = 0;
	int n;
	int cdblen;
	static unsigned char buf[258];

	// Process any incoming messages from the initiator
	while (adapter_GetScsiBits() & SCSI_BIT_ATN)
		{
		n = scsi_ReadMessageFromHost(buf,sizeof(buf));
		if (n <= 0)
			{
			adapter_Disconnect();
			return;
			}
		if (buf[0] & 0x80)					// Identify message?
			{
			identify_msg = buf[0];
			lun = identify_msg & 0x07;
			}
		else
			{
			switch (buf[0])
				{
				case 0x00:		// command complete
				case 0x07:		// message reject
				case 0x08:		// nop
				case 0x09:		// message parity error
					// Ignore these here
					break;

				case 0x05:		// initiator detected error
				case 0x06:		// abort
					adapter_Disconnect();
					return;

				case 0x0C:		// bus device reset
					// Log it and abort the transaction
					tspi_DeviceReset();
					adapter_Disconnect();
					return;

				default:
					// Reject any other messages
					buf[0] = 0x07;		// message reject
					if (adapter_Write(SCSI_PHASE_MSG_IN,&buf[0],1) != 1)
						{
						adapter_Disconnect();
						return;
						}
				}
			}
		}


	// Now get the CDB command byte from the initiator
	if (adapter_Read(SCSI_PHASE_CMD,&buf[0],1) != 1)
		{
		adapter_Disconnect();
		return;
		}

	// Determine the CDB length
	switch (buf[0] >> 5)
		{
		case 0:		cdblen = 6;		break;
		case 1:		cdblen = 10;	break;
		case 2:		cdblen = 10;	break;
		case 5:		cdblen = 12;	break;
		case 6:		cdblen = 6;		break;
		case 7:		cdblen = 10;	break;
		default:	cdblen = 1;		break;
		}

	// Get the rest of the CDB
	if (cdblen > 1)
		{
		n = cdblen - 1;
		if (adapter_Read(SCSI_PHASE_CMD,&buf[1],n) != n)
			{
			adapter_Disconnect();
			return;
			}
		}


	tspi_ProcessIncomingEvent(initiator_id,lun,identify_msg,cdblen,buf);
	}



int scsi_Init(char FAR *cmdline)
	{
	return adapter_Init(cmdline);
	}



void scsi_Term()
	{
	adapter_Term();
	}

void scsi_Reset()
	{
	adapter_Reset();
	}

