
#include <limits.h>

#include "tspi.h"

#include "scsi.h"
#include "tspilink.h"
#include "tspiprev.h"
#include "irq.h"
#include "scsidefs.h"

#define TSPI_VERSION	0x00000104L

// Flag bits used internally
#define TSPI_FLAG_InUse			0x8000
#define TSPI_FLAG_Processing	0x4000
#define TSPI_FLAG_LunAttached	0x2000


#define NUM_LUNS	8


TSPI_EVENT		TspiEvent[NUM_LUNS];		// temporary CDB storage
int 			TspiFlags[NUM_LUNS];		// Internal flags
unsigned char	TspiIdentify[NUM_LUNS];		// Used for reconnect
unsigned char	AdapterNum;					// Our TSPI adapter index


// We need a pointer to any previously installed TSPI drivers
// so we can pass requests down to them.

int (_cdecl FAR *tspi_PrevEntry)(void FAR *p);



// We need some default sense and inquiry data for LUNs that
// aren't handled by an application

#define INVALID_LUN_SENSE_LEN	18
unsigned char InvalidLunSenseData[INVALID_LUN_SENSE_LEN] =
	{
	0x70,					// Current error
	0x00,					// Segment
	SCSI_Skey_NotReady,		// Sense key
	0x00, 0x00, 0x00, 0x00,
	INVALID_LUN_SENSE_LEN - 7,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

#define INVALID_LUN_INQUIRY_LEN	36
unsigned char InvalidLunInquiryData[INVALID_LUN_INQUIRY_LEN] =
	{
	0x3F,			// unknown device, not attached yet
	0x00,			// device qualifier 
	0x00,			// ANSI version 2 
	0x00,			// Response format 2
	0x20,			// 32 additional bytes
	0x00,			// Reserved
	0x00,			// Reserved
	0x00,			// Attributes (wide, sync, etc.)
	'T',			// Vendor ID ("TSPI    ")
	'S', 'P', 'I', ' ', ' ', ' ', ' ',
	'N',			// Product ID ("NOT-ATTACHED    ");
	'O', 'T', '-', 'A', 'T', 'T', 'A', 'C', 'H', 'E', 'D', ' ', ' ', ' ', ' ',
	'1',			// Product Revision ("1.04")
	'.', '0', '4'
	};




// This routine is called by SCSI.C whenever a SCSI BUS RESET is detected
// We set a flag bit for each LUN so we can notify them on the next command

void tspi_BusReset(void)
	{
	int i;
	for (i=0; i<NUM_LUNS; i++)
		TspiFlags[i] |= TSPI_FLAG_BusReset;
	}


// This routine is called by SCSI.C whenever a BUS DEVICE RESET message arrives 
// We set a flag bit for each LUN so we can notify them on the next command

void tspi_DeviceReset(void)
	{
	int i;
	for (i=0; i<NUM_LUNS; i++)
		TspiFlags[i] |= TSPI_FLAG_DeviceReset;
	}


// This routine is called by SCSI.C whenever a CDB is received
// If an application has attached to that LUN we log the information 
// in the TspiEvent[] array.  If not, we implement default INQUIRY, 
// REQUEST SENSE, and TEST UNIT READY commands for the invalid LUN.

void tspi_ProcessIncomingEvent(	int					initiator_id,
								int					lun,
								unsigned char		identify_msg,
								int					cdblen,
								unsigned char FAR *cdbbyte )
	{
	int n;

	// Make sure we aren't still processing the last command
	if (TspiFlags[lun] & TSPI_FLAG_InUse)
		{
		// Send BUSY status back to the initiator
		scsi_CompleteCommand(SCSI_Status_Busy);
		return;
		}

	// Save the identify message for later reconnects
	TspiIdentify[lun] = identify_msg;


	// Store the command information in the TspiEvent array
	TspiEvent[lun].InitiatorId = initiator_id;
	TspiEvent[lun].Lun = lun;
	TspiEvent[lun].Flags = 0;
	if (identify_msg & 0x40)
		{
		TspiEvent[lun].Flags |= TSPI_FLAG_DiscPriv;
		TspiFlags[lun] |= TSPI_FLAG_DiscPriv;
		}
	if (identify_msg & 0x20)
		{
		TspiEvent[lun].Flags |= TSPI_FLAG_TargetRoutine;
		TspiFlags[lun] |= TSPI_FLAG_TargetRoutine;
		}
	TspiEvent[lun].CdbLength = cdblen;
	for (n=0; n<cdblen; n++)
		TspiEvent[lun].CdbByte[n] = *(cdbbyte+n);


	// If the LUN isn't attached we have to handle the CDB ourselves
	if ((TspiFlags[lun] & TSPI_FLAG_LunAttached) == 0)
		{
		if (TspiEvent[lun].CdbLength > 0)	// Any CDB bytes received?
			{
			switch (TspiEvent[lun].CdbByte[0])
				{
				case SCSI_Cmd_RequestSense:
					n = TspiEvent[lun].CdbByte[4];
					if (n > INVALID_LUN_SENSE_LEN)
						n = INVALID_LUN_SENSE_LEN;
					if (n)
						scsi_WriteToHost(&InvalidLunSenseData[0],n);
					scsi_CompleteCommand(SCSI_Status_Good);
					break;

#if 0
				case SCSI_Cmd_Inquiry:
					n = TspiEvent[lun].CdbByte[4];
					if (n > INVALID_LUN_INQUIRY_LEN)
						n = INVALID_LUN_INQUIRY_LEN;
					if (n)
						scsi_WriteToHost(&InvalidLunInquiryData[0],n);
					scsi_CompleteCommand(SCSI_Status_Good);
					break;

				case SCSI_Cmd_TestUnitReady:
					// Not ready
					scsi_CompleteCommand(SCSI_Status_CheckCondition);
					break;
#endif

				default:
					scsi_CompleteCommand(SCSI_Status_CheckCondition);
				}
			}
		else
			scsi_CompleteCommand(SCSI_Status_Good);
		}
	else
		{
		// Flag the event for application processing
		TspiFlags[lun] |= TSPI_FLAG_InUse;
		}
	}





// This routine waits for an incoming TSPI_EVENT record.  If one
// arrives it is copied to the user buffer.

void WaitForEvent(TSPI_EVENT FAR *p)
	{
	long ticks_to_wait = p->Timeout ? (p->Timeout / MS_PER_TICK) + 1 : 0;
	long start_time = GetTickCount();
	int i;

	// Wait for a CDB or reset event to arrive
	while ((TspiFlags[p->Lun] & (TSPI_FLAG_InUse | TSPI_FLAG_BusReset | TSPI_FLAG_DeviceReset)) == 0)
		{
		if ((GetTickCount() - start_time) >= ticks_to_wait)
			{
			p->Error = TSPI_ERR_Timeout;
			return;
			}
		scsi_Poll();
		}

	// Make sure we aren't already processing this one
	if (TspiFlags[p->Lun] & TSPI_FLAG_Processing)
		{
		p->Error = TSPI_ERR_InvalidCommand;
		return;
		}

	// We have a new event to return. 

	// Copy any pending reset flags to the user buffer
	_disable();
	p->Flags = TspiFlags[p->Lun] & ~(TSPI_FLAG_InUse | TSPI_FLAG_Processing | TSPI_FLAG_LunAttached);
	TspiFlags[p->Lun] &= ~(TSPI_FLAG_BusReset | TSPI_FLAG_DeviceReset);
	_enable();

	// OR in the remaining flags
	p->Flags |= TspiEvent[p->Lun].Flags;

	// Copy the incoming CDB to the user buffer (if one is available)
	if (TspiFlags[p->Lun] & TSPI_FLAG_InUse)
		{
		p->CdbLength = TspiEvent[p->Lun].CdbLength;
		for (i=0; i<p->CdbLength; i++)
			p->CdbByte[i] = TspiEvent[p->Lun].CdbByte[i];
		TspiFlags[p->Lun] |= TSPI_FLAG_Processing;
		}
	else
		{
		p->CdbLength = 0;		// No CDB, just the reset flags
		}

	// Fill remaining CDB bytes with 0
	for (i=p->CdbLength; i<sizeof(p->CdbByte); i++)
		p->CdbByte[i] = 0;

	p->Error = TSPI_ERR_None;
	}


// This routine completes a SCSI command, sending the command status
// back to the initiator and disconnecting from the SCSI bus
// It also releases control of the TspiEvent structure

void CompleteCommand(TSPI_CMD FAR *p)
	{
	TspiFlags[p->Lun] &= ~(TSPI_FLAG_InUse | TSPI_FLAG_Processing);
	scsi_CompleteCommand((unsigned char)(p->Parm[0]));
	p->Error = TSPI_ERR_None;
	}


// This routine transfers data from the initiator into the user buffer
// It breaks large transfers up into smaller chunks to work around the
// 64K segment limit of the x86 real-mode architecture.

void ReadFromHost(TSPI_XFER FAR *p)
	{
	char _huge *bufp = (char _huge *)p->TransferAddress;
	int n,m;

	p->Error = TSPI_ERR_None;
	p->ResidualLength = p->TransferLength;
	while (p->ResidualLength > 0)
		{
		if (p->ResidualLength < 16384)
			m = p->ResidualLength;
		else
			m = 16384;
		n = scsi_ReadFromHost(bufp,m);
		if (n < 0)
			p->ResidualLength -= -n;
		else
			p->ResidualLength -= n;
		if (n != m)
			{
			p->Error = TSPI_ERR_Initiator;
			break;
			}
		bufp += n;
		}
	}


// This routine transfers data from the user buffer to the initiator.
// It breaks large transfers up into smaller chunks to work around the
// 64K segment limit of the x86 real-mode architecture.

void WriteToHost(TSPI_XFER FAR *p)
	{
	char _huge *bufp = (char _huge *)p->TransferAddress;
	int n,m;

	p->Error = TSPI_ERR_None;
	p->ResidualLength = p->TransferLength;
	while (p->ResidualLength > 0)
		{
		if (p->ResidualLength < 16384)
			m = p->ResidualLength;
		else
			m = 16384;
		n = scsi_WriteToHost(bufp,m);
		if (n < 0)
			p->ResidualLength -= -n;
		else
			p->ResidualLength -= n;
		if (n != m)
			{
			p->Error = TSPI_ERR_Initiator;
			break;
			}
		bufp += n;
		}
	}


// This routine transfers message bytes from the user buffer to the
// initiator.  The transfer is limited to 258 bytes, which is the
// size of the largest possible message.  Make multiple calls to
// send multiple messages.

void SendMessage(TSPI_XFER FAR *p)
	{
	int n,m;

	p->Error = TSPI_ERR_None;
	p->ResidualLength = p->TransferLength;

	if (p->TransferLength > 258)		
		{
		p->Error = TSPI_ERR_TooLarge;
		return;
		}

	m = (int) p->TransferLength;

	n = scsi_WriteMessageToHost(p->TransferAddress,m);

	if (n < 0)
		p->ResidualLength -= -n;
	else
		p->ResidualLength -= n;

	if (n != m)
		p->Error = TSPI_ERR_Initiator;
	}


// This routine transfers a single SCSI message from the initiator into
// the user buffer.  The ResidualLength field can be used to determine
// the length of the message.

void GetMessage(TSPI_XFER FAR *p)
	{
	int n;
	int m;

	p->Error = TSPI_ERR_None;
	p->ResidualLength = p->TransferLength;

	if (p->TransferLength > INT_MAX)
		m = INT_MAX;
	else
		m = (int) p->TransferLength;

	n = scsi_ReadMessageFromHost(p->TransferAddress,m);

	if (n < 0)
		p->ResidualLength -= -n;
	else
		p->ResidualLength -= n;

	if (n != m)
		p->Error = TSPI_ERR_Initiator;
	}


// This routine disconnects us from the SCSI bus, possibly sending
// a SAVE DATA POINTER message to the host.

void Disconnect(TSPI_CMD FAR *p)
	{
	if (TspiFlags[p->Lun] & TSPI_FLAG_DiscPriv)
		{
		if (scsi_Disconnect( (p->Flags & TSPI_FLAG_SaveDataPointer) ? 1 : 0 ) )
			p->Error = TSPI_ERR_None;
		else
			p->Error = TSPI_ERR_Initiator;
		}
	else
		p->Error = TSPI_ERR_InvalidCommand;
	}


// This routine reconnects us to the specified initiator.  It will also
// send the same IDENTIFY back to the host to identify the I_T_L nexus.

void Reconnect(TSPI_CMD FAR *p)
	{
	if (TspiFlags[p->Lun] & TSPI_FLAG_DiscPriv)
		{
		if (scsi_Reconnect(p->InitiatorId,TspiIdentify[p->Lun]))
			p->Error = TSPI_ERR_None;
		else
			p->Error = TSPI_ERR_Initiator;
		}
	else
		p->Error = TSPI_ERR_InvalidCommand;
	}


// This routine is the main TSPI Manager entry point.  All user requests
// come here for processing.  We handle requests for our adapter index, and
// route all other requests down to any previously installed TSPI drivers

int FAR _loadds tspi_Entry(TSPI_CMD FAR *p)
	{
	// See if this command is meant for our adapter
	if (p->AdapterIndex < AdapterNum)
		{
		// Send it to a previously loaded TSPI manager (if available)
		if (tspi_PrevEntry)
			tspi_PrevEntry(p);
		else
			p->Error = TSPI_ERR_InvalidAdapter;

		// Update the adapter count if this is a TSPI_CMD_AdapterInfo command
		if (p->CommandCode == TSPI_CMD_AdapterInfo)
			p->Parm[0] = AdapterNum + 1;

		return p->Error;
		}

	if (p->AdapterIndex > AdapterNum)
		{
		p->Error = TSPI_ERR_InvalidAdapter;
		return p->Error;
		}


	// Process the TSPI command
	switch (p->CommandCode)
		{
		case TSPI_CMD_AdapterInfo:
			p->Parm[0] = AdapterNum + 1;		// adapter count
			p->Parm[1] = TSPI_VERSION;			// our version
			p->Parm[2] = scsi_GetScsiId();		// our target ID
			p->Parm[3] = 0;						// reserved
			p->Parm[4] = 0;						// reserved
			p->Parm[5] = 0;						// reserved
			p->Error = TSPI_ERR_None;
			break;

		case TSPI_CMD_AttachLUN:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					p->Error = TSPI_ERR_LunNotAvailable;
				else
					{
					TspiFlags[p->Lun] = TSPI_FLAG_LunAttached;
					scsi_Reset();
					p->Error = TSPI_ERR_None;
					}
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_DetachLUN:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					TspiFlags[p->Lun] = 0;
					p->Error = TSPI_ERR_None;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_GetEvent:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					WaitForEvent((TSPI_EVENT FAR *)p);
					if (scsi_IsAtnAsserted())
						p->Flags |= TSPI_FLAG_HostMsgWaiting;
					else
						p->Flags &= ~TSPI_FLAG_HostMsgWaiting;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;


		case TSPI_CMD_CompleteCommand:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					CompleteCommand(p);
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_ReadFromHost:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					ReadFromHost((TSPI_XFER FAR *)p);
					if (scsi_IsAtnAsserted())
						p->Flags |= TSPI_FLAG_HostMsgWaiting;
					else
						p->Flags &= ~TSPI_FLAG_HostMsgWaiting;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_WriteToHost:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					WriteToHost((TSPI_XFER FAR *)p);
					if (scsi_IsAtnAsserted())
						p->Flags |= TSPI_FLAG_HostMsgWaiting;
					else
						p->Flags &= ~TSPI_FLAG_HostMsgWaiting;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_SendMessage:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					SendMessage((TSPI_XFER FAR *)p);
					if (scsi_IsAtnAsserted())
						p->Flags |= TSPI_FLAG_HostMsgWaiting;
					else
						p->Flags &= ~TSPI_FLAG_HostMsgWaiting;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_GetMessage:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					GetMessage((TSPI_XFER FAR *)p);
					if (scsi_IsAtnAsserted())
						p->Flags |= TSPI_FLAG_HostMsgWaiting;
					else
						p->Flags &= ~TSPI_FLAG_HostMsgWaiting;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_Disconnect:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					Disconnect(p);
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		case TSPI_CMD_Reconnect:
			if (p->Lun < NUM_LUNS)
				{
				if (TspiFlags[p->Lun] & TSPI_FLAG_LunAttached)
					{
					Reconnect(p);
					if (scsi_IsAtnAsserted())
						p->Flags |= TSPI_FLAG_HostMsgWaiting;
					else
						p->Flags &= ~TSPI_FLAG_HostMsgWaiting;
					}
				else
					p->Error = TSPI_ERR_LunNotAvailable;
				}
			else
				p->Error = TSPI_ERR_InvalidLUN;
			break;

		default:
			p->Error = TSPI_ERR_InvalidCommand;
			break;
		}

	return p->Error;
	}


// This routine is called to initialize the TSPI manager.  It calls
// the scsi_Init() routine to initialize the SCSI subsystem.

int tspi_Init(char FAR *cmdline)
	{
	int i;

	for (i=0; i<NUM_LUNS; i++)
		TspiFlags[i] = 0;

	AdapterNum = 0;

	tspi_PrevEntry = GetPrevTspiEntry();
	if (tspi_PrevEntry)
		{
		TSPI_CMD cmd;
		// Set our adapter number based on the 
		// number of previously installed adapters
		cmd.CommandCode = TSPI_CMD_AdapterInfo;
		cmd.Error = -1;
		cmd.Flags = 0;
		cmd.AdapterIndex = 0;
		tspi_PrevEntry(&cmd);
		if (cmd.Error == TSPI_ERR_None)
			{
			AdapterNum = cmd.Parm[0];
			PutString("Adapter = ");
			PutHexWord(AdapterNum);
			PutString("\r\n");
			}
		else
			tspi_PrevEntry = 0;
		}

	return scsi_Init(cmdline);
	}

void tspi_Term()
	{
	scsi_Term();
	}
