/* ***********************************************************************
	PCI_8XX.C - Symbios 53C8XX PCI Utility functions
	Author: Andy Brown - Systems Engineering
	Date: 08/23/95

	Purpose:  This code provides the following standard utility functions
	for manipulating and using Symbios 53C8XX pci devices.
		Functions provided:
			C8XX_DevID2Str - Converts a given 8xx device ID to a string descriptor
			C8XX_BuildDeviceList - Finds all 53C8XX devices in a system and places
			device info into a standard data structure
			C8XX_ChooseDevice - allows user to select a specific PCI device to
			work with
			C8XX_GetMemBase - Gets the memory base address of the specified 8xx
			device
			C8XX_GetIOBase - Gets the IO base address of the specified 8xx
			device
			C8XX_GetRAMBase - Gets the SCRIPT RAM base address of the specified 8xx
			device

	Note: This code requires that PCI.C be linked in with it as it uses the
			functions in that code to do low level accesses to PCI config space

	Revision History:
	1.0.0 - Initial Revision - ACB  08/23/95

	1.0.1 - SLR  02/14/96
			  Fixed problem with C8XX_BuildDeviceList when multiple controllers
			  with the same Symbios chip are in the system

	1.0.2 - SLR  05/28/96
			  Added supprot for the 53C895

************************************************************************** */

/* the following two #includes must go in this order */
#include "pci.h"				/* low level pci functions */
#include "pci_8xx.h"
#include <stdio.h>

/* Local Global Variables */
static unsigned int ID_Array[] = { 			/* used to scan for 8XX PCI devices */
	C810_DEVICE_ID, C815_DEVICE_ID,
	C820_DEVICE_ID, C825_DEVICE_ID,
	C810AP_DEVICE_ID, C860_DEVICE_ID,
	C875_DEVICE_ID, C895_DEVICE_ID, NULL_DEVICE_ID };

/* Function Prototypes */
char *C8XX_DevID2Str(int id, unsigned char rev_id);
int C8XX_BuildDeviceList(device_record *Device[], int Max_Devices);
int C8XX_ChooseDevice(device_record *Device[], int NumDevices);
unsigned int C8XX_GetMemBase(device_record *Device);
unsigned int C8XX_GetIOBase(device_record *Device);
unsigned int C8XX_GetRAMBase(device_record *Device);

/******************************************************************************/
/* Function: C8XX_DevID2Str
	Purpose : Returns 53C8XX Device Name string given a 53C8XX Device ID
	Input:  Valid 53C8XX device ID
	Output:	Descriptor string of 53C8xx device
				"Unknown Symbios SCSI Device" if the ID pass in is not known
	Assumptions: None
	Other functions called: None
*/
/******************************************************************************/

char *C8XX_DevID2Str(int id, unsigned char rev_id)
{

	switch (id) {
		case C810_DEVICE_ID: {
			/* Could be 810 or 810a, check revID register to tell */
			if ((rev_id & 0xF0) == C810_VERSION_ID) return("SYM 53C810");
			if ((rev_id & 0xF0) == C810A_VERSION_ID) return("SYM 53C810A");
			if ((rev_id & 0xF0) == C810ALV_VERSION_ID) return("SYM 53C810ALV");
		}
		case C815_DEVICE_ID:
			return("SYM 53C815");
		case C820_DEVICE_ID:
			return("SYM 53C820");
		case C825_DEVICE_ID: {
			/* Could be 825 or 825a, check revID register to tell */
			if ((rev_id & 0xF0) == C825_VERSION_ID) return("SYM 53C825");
			if ((rev_id & 0xF0) == C825A_VERSION_ID) return("SYM 53C825A");
		}
		case C810AP_DEVICE_ID:
			return("SYM 53C810A Proto");
		case C860_DEVICE_ID:
			if ((rev_id & 0xF0) == C860_VERSION_ID) return("SYM 53C860");
			if ((rev_id & 0xF0) == C860LV_VERSION_ID) return("SYM 53C860LV");
		case C875_DEVICE_ID:
			return("SYM 53C875");
		case C895_DEVICE_ID:
			return("SYM 53C895");
		default:
			return("Unknown Symbios SCSI Device");
	}
}

/******************************************************************************/
/* Function: BuildDeviceList
	Purpose : Builds a list of 53C8XX devices found in the system
	Input:  Pointer to an array of device_record structures and a limit
			  Max_Devices indicating how big the device_record array is.
	Output:	This function fills the Device data structure by calling
				the PCI_FindSYMDevice function for all possible ID's.
				If Max_Devices is reached then the function returns immediately
				with a return value = Max_Devices, otherwise it returns the
				number of devices found.
	Assumptions: That enough Device structures have been allocated to fill in
					 Max_Devices structures
	Other functions called: PCI_FindSYMDevice
*/
/******************************************************************************/

int C8XX_BuildDeviceList(device_record *Device[], int Max_Devices)

{
	int NumDevices;			/* Actual number of devices found */
	int i;						/* Counting variable */

	if (!Max_Devices) {		/* Ooops!!! didn't allocate any device space */
		return(0);
	}

	NumDevices = 0;
	i = 0;

	/* look for devices */
	while (ID_Array[i] != NULL_DEVICE_ID) {
		Device[NumDevices]->device_id = ID_Array[i];
		Device[NumDevices]->device_index = 0;
		/* find all devices with a particular device id */
		while (PCI_FindSYMDevice(Device[NumDevices]) != PCI_NO_DEVICE_FOUND) {
			NumDevices++;
			/* did we run out of space to store devices, if yes return */
			if (NumDevices == Max_Devices) return(NumDevices+1);
			Device[NumDevices]->device_index = Device[NumDevices-1]->device_index+1;
			Device[NumDevices]->device_id = Device[NumDevices-1]->device_id;
		}
		i++;
	}

	return(NumDevices);

}

/******************************************************************************/
/* Function: C8XX_ChooseDevice
	Purpose : Allows the selection of a particular 8xx device in the system
				 Displays all devices and prompts for a selection
	Input:  Array of device_records (for all devices in the system) and the
			  actual number of devices in the system
	Output:	Returns the index of the selected device used to index into the
				Device array
	Assumptions: That NumDevices gives the proper number of devices in the
					 Device array
	Other functions called: C8XX_DeviceID2Str
*/
/******************************************************************************/

int C8XX_ChooseDevice(device_record *Device[], int NumDevices)

{
	int i;
	int device_select = -1;

	/* print devices from the Device array */
	for (i=0; i<NumDevices; i++) {
		printf("%d: %s Bus #:%X Device #:%X Function :%X\n", i,
							C8XX_DevID2Str(Device[i]->device_id, Device[i]->rev_id),
							Device[i]->bus_num, Device[i]->device_num>>3,
							Device[i]->function);
	}
	printf("\n");
	/* prompt for device selection */
	while (device_select == -1) {
		printf("Choose a device: ");
		scanf("%d", &device_select);
		fflush(stdin);
		if ((device_select < 0) || (device_select > NumDevices-1)) {
			device_select = -1;
		}
	}

	return(device_select);
}

/******************************************************************************/
/* Function: GetMemBase
	Purpose : Fills in the Device structure with the Memory Base address
				 of the 53C8XX device
	Input: Pointer to Structure device_record Device which indicates which
			 device to get the Memory base address for
	Output: Fills in the Device->base_addr1 field and returns C8XX_GOOD status
			  if the value read is non-zero.  Returns C8XX_NO_BASE if the value
			  read is zero
	Assumptions: Pointer to device structure is valid.
	Other functions called: PCI_GetConfigRegister
*/
/******************************************************************************/

unsigned int C8XX_GetMemBase(device_record *Device)
{

	/* Read config register that contains the Register Memory Base address */
	Device->base_addr1 = PCI_GetConfigRegister(Device, C8XX_CONFIG_REG_MEMB);

	if(Device->base_addr1 & 0xFFFFFFFCL) return(C8XX_GOOD);
	else return(C8XX_NO_BASE); /* mem base address was not programed by system */

}

/******************************************************************************/
/* Function: GetIOBase
	Purpose : Fills in the Device structure with the IO Base address
				 of the 53C8XX device
	Input: Pointer to Structure device_record Device which indicates which
			 device to get the IO base address for
	Output: Fills in the Device->base_addr2 field and returns C8XX_GOOD status
			  if the value read is non-zero.  Returns C8XX_NO_BASE if the value
			  read is zero
	Assumptions: Pointer to device structure is valid.
	Other functions called: PCI_GetConfigRegister
*/
/******************************************************************************/

unsigned int C8XX_GetIOBase(device_record *Device)
{
	/* Read config register that contains the Register IO Base address */
	Device->base_addr2 = PCI_GetConfigRegister(Device, C8XX_CONFIG_REG_IOB);

	if(Device->base_addr2 & 0xFFFFFFFCL) return(C8XX_GOOD);
	else return(C8XX_NO_BASE); /* IO base address was not programmed by system */

}

/******************************************************************************/
/* Function: GetRAMBase
	Purpose : Fills in the Device structure with the SCRIPT RAM Base address
				 of the 53C825A or 875 device
	Input: Pointer to Structure device_record Device which indicates which
			 device to get the SCRIPT RAM base address for
	Output: Fills in the Device->base_addr3 field and returns C8XX_GOOD status
			  if the value read is non-zero.  Returns C8XX_NO_BASE if the value
			  read is zero
	Assumptions: Pointer to device structure is valid and the device is a 53C825A
					 or a 53C875.
	Other functions called: PCI_GetConfigRegister
*/
/******************************************************************************/

unsigned int C8XX_GetRAMBase(device_record *Device)
{
	/* Read config register that contains the SCRIPT RAM Base address */
	Device->base_addr3 = PCI_GetConfigRegister(Device, C8XX_CONFIG_REG_RAMB);

	if(Device->base_addr3 & 0xFFFFFFFCL) return(C8XX_GOOD);
	else return(C8XX_NO_BASE); /* RAM base address was not programmed by system */

}
